@@ -3,22 +3,25 @@
  * parse_func.c
  *		handle function calls in parser
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $PostgreSQL: pgsql/src/backend/parser/parse_func.c,v 1.182 2005/10/15 02:49:22 momjian Exp $
+ *	  $PostgreSQL: pgsql/src/backend/parser/parse_func.c,v 1.201.2.1 2010/07/30 17:57:07 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
 #include "access/heapam.h"
+#include "catalog/pg_attrdef.h"
+#include "catalog/pg_constraint.h"
 #include "catalog/pg_inherits.h"
 #include "catalog/pg_proc.h"
+#include "catalog/pg_type.h"
 #include "funcapi.h"
-#include "lib/stringinfo.h"
+#include "miscadmin.h"
 #include "nodes/makefuncs.h"
 #include "parser/parse_agg.h"
 #include "parser/parse_coerce.h"
@@ -27,15 +30,19 @@
 #include "parser/parse_relation.h"
 #include "parser/parse_target.h"
 #include "parser/parse_type.h"
+#include "parser/parsetree.h"
 #include "utils/builtins.h"
 #include "utils/fmgroids.h"
 #include "utils/lsyscache.h"
 #include "utils/syscache.h"
 
 
+static Oid	FuncNameAsType(List *funcname);
 static Node *ParseComplexProjection(ParseState *pstate, char *funcname,
-					   Node *first_arg);
-static void unknown_attribute(ParseState *pstate, Node *relref, char *attname);
+					   Node *first_arg, int location);
+static void unknown_attribute(ParseState *pstate, Node *relref, char *attname,
+				  int location);
+static bool check_pg_get_expr_arg(ParseState *pstate, Node *arg, int netlevelsup);
 
 
 /*
@@ -59,7 +66,8 @@ static void unknown_attribute(ParseState *pstate, Node *relref, char *attname);
  */
 Node *
 ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
-				  bool agg_star, bool agg_distinct, bool is_column)
+				  bool agg_star, bool agg_distinct, bool is_column,
+				  int location)
 {
 	Oid			rettype;
 	Oid			funcid;
@@ -83,16 +91,17 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 		ereport(ERROR,
 				(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
 				 errmsg("cannot pass more than %d arguments to a function",
-						FUNC_MAX_ARGS)));
+						FUNC_MAX_ARGS),
+				 parser_errposition(pstate, location)));
 
 	/*
 	 * Extract arg type info in preparation for function lookup.
 	 *
-	 * If any arguments are Param markers of type VOID, we discard them from the
-	 * parameter list.	This is a hack to allow the JDBC driver to not have to
-	 * distinguish "input" and "output" parameter symbols while parsing
-	 * function-call constructs.  We can't use foreach() because we may modify
-	 * the list ...
+	 * If any arguments are Param markers of type VOID, we discard them from
+	 * the parameter list.	This is a hack to allow the JDBC driver to not
+	 * have to distinguish "input" and "output" parameter symbols while
+	 * parsing function-call constructs.  We can't use foreach() because we
+	 * may modify the list ...
 	 */
 	nargs = 0;
 	for (l = list_head(fargs); l != NULL; l = nextl)
@@ -131,7 +140,8 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 		{
 			retval = ParseComplexProjection(pstate,
 											strVal(linitial(funcname)),
-											first_arg);
+											first_arg,
+											location);
 			if (retval)
 				return retval;
 
@@ -156,7 +166,7 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 	if (fdresult == FUNCDETAIL_COERCION)
 	{
 		/*
-		 * We can do it as a trivial coercion. coerce_type can handle these
+		 * We interpreted it as a type coercion. coerce_type can handle these
 		 * cases, so why duplicate code...
 		 */
 		return coerce_type(pstate, linitial(fargs),
@@ -174,12 +184,14 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
 			   errmsg("%s(*) specified, but %s is not an aggregate function",
 					  NameListToString(funcname),
-					  NameListToString(funcname))));
+					  NameListToString(funcname)),
+					 parser_errposition(pstate, location)));
 		if (agg_distinct)
 			ereport(ERROR,
 					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
 			errmsg("DISTINCT specified, but %s is not an aggregate function",
-				   NameListToString(funcname))));
+				   NameListToString(funcname)),
+					 parser_errposition(pstate, location)));
 	}
 	else if (fdresult != FUNCDETAIL_AGGREGATE)
 	{
@@ -193,7 +205,8 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 		{
 			Assert(nargs == 1);
 			Assert(list_length(funcname) == 1);
-			unknown_attribute(pstate, first_arg, strVal(linitial(funcname)));
+			unknown_attribute(pstate, first_arg, strVal(linitial(funcname)),
+							  location);
 		}
 
 		/*
@@ -206,7 +219,8 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 							func_signature_string(funcname, nargs,
 												  actual_arg_types)),
 					 errhint("Could not choose a best candidate function. "
-							 "You may need to add explicit type casts.")));
+							 "You might need to add explicit type casts."),
+					 parser_errposition(pstate, location)));
 		else
 			ereport(ERROR,
 					(errcode(ERRCODE_UNDEFINED_FUNCTION),
@@ -214,18 +228,20 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 							func_signature_string(funcname, nargs,
 												  actual_arg_types)),
 			errhint("No function matches the given name and argument types. "
-					"You may need to add explicit type casts.")));
+					"You might need to add explicit type casts."),
+					 parser_errposition(pstate, location)));
 	}
 
 	/*
-	 * enforce consistency with ANYARRAY and ANYELEMENT argument and return
-	 * types, possibly adjusting return type or declared_arg_types (which will
-	 * be used as the cast destination by make_fn_arguments)
+	 * enforce consistency with polymorphic argument and return types,
+	 * possibly adjusting return type or declared_arg_types (which will be
+	 * used as the cast destination by make_fn_arguments)
 	 */
 	rettype = enforce_generic_type_consistency(actual_arg_types,
 											   declared_arg_types,
 											   nargs,
-											   rettype);
+											   rettype,
+											   false);
 
 	/* perform the necessary typecasting of arguments */
 	make_fn_arguments(pstate, fargs, actual_arg_types, declared_arg_types);
@@ -250,10 +266,21 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 
 		aggref->aggfnoid = funcid;
 		aggref->aggtype = rettype;
-		aggref->target = linitial(fargs);
+		aggref->args = fargs;
 		aggref->aggstar = agg_star;
 		aggref->aggdistinct = agg_distinct;
 
+		/*
+		 * Reject attempt to call a parameterless aggregate without (*)
+		 * syntax.	This is mere pedantry but some folks insisted ...
+		 */
+		if (fargs == NIL && !agg_star)
+			ereport(ERROR,
+					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+					 errmsg("%s(*) must be used to call a parameterless aggregate function",
+							NameListToString(funcname)),
+					 parser_errposition(pstate, location)));
+
 		/* parse_agg.c does additional aggregate-specific processing */
 		transformAggregateCall(pstate, aggref);
 
@@ -262,9 +289,13 @@ ParseFuncOrColumn(ParseState *pstate, List *funcname, List *fargs,
 		if (retset)
 			ereport(ERROR,
 					(errcode(ERRCODE_INVALID_FUNCTION_DEFINITION),
-					 errmsg("aggregates may not return sets")));
+					 errmsg("aggregates cannot return sets"),
+					 parser_errposition(pstate, location)));
 	}
 
+	/* Hack to protect pg_get_expr() against misuse */
+	check_pg_get_expr_args(pstate, funcid, fargs);
+
 	return retval;
 }
 
@@ -502,10 +533,10 @@ func_select_candidate(int nargs,
 	/*
 	 * Still too many candidates? Try assigning types for the unknown columns.
 	 *
-	 * NOTE: for a binary operator with one unknown and one non-unknown input, we
-	 * already tried the heuristic of looking for a candidate with the known
-	 * input type on both sides (see binary_oper_exact()). That's essentially
-	 * a special case of the general algorithm we try next.
+	 * NOTE: for a binary operator with one unknown and one non-unknown input,
+	 * we already tried the heuristic of looking for a candidate with the
+	 * known input type on both sides (see binary_oper_exact()). That's
+	 * essentially a special case of the general algorithm we try next.
 	 *
 	 * We do this by examining each unknown argument position to see if we can
 	 * determine a "type category" for it.	If any candidate has an input
@@ -518,10 +549,10 @@ func_select_candidate(int nargs,
 	 * If we are able to determine a type category, also notice whether any of
 	 * the candidates takes a preferred datatype within the category.
 	 *
-	 * Having completed this examination, remove candidates that accept the wrong
-	 * category at any unknown position.  Also, if at least one candidate
-	 * accepted a preferred type at a position, remove candidates that accept
-	 * non-preferred types.
+	 * Having completed this examination, remove candidates that accept the
+	 * wrong category at any unknown position.	Also, if at least one
+	 * candidate accepted a preferred type at a position, remove candidates
+	 * that accept non-preferred types.
 	 *
 	 * If we are down to one candidate at the end, we win.
 	 */
@@ -648,7 +679,7 @@ func_select_candidate(int nargs,
  *	(exact match) is as quick as possible.
  *
  * If an exact match isn't found:
- *	1) check for possible interpretation as a trivial type coercion
+ *	1) check for possible interpretation as a type coercion request
  *	2) get a vector of all possible input arg type arrays constructed
  *	   from the superclasses of the original input arg types
  *	3) get a list of all possible argument type arrays to the function
@@ -699,25 +730,31 @@ func_get_detail(List *funcname,
 		 * If we didn't find an exact match, next consider the possibility
 		 * that this is really a type-coercion request: a single-argument
 		 * function call where the function name is a type name.  If so, and
-		 * if we can do the coercion trivially (no run-time function call
-		 * needed), then go ahead and treat the "function call" as a coercion.
+		 * if the coercion path is RELABELTYPE or COERCEVIAIO, then go ahead
+		 * and treat the "function call" as a coercion.
+		 *
 		 * This interpretation needs to be given higher priority than
 		 * interpretations involving a type coercion followed by a function
 		 * call, otherwise we can produce surprising results. For example, we
-		 * want "text(varchar)" to be interpreted as a trivial coercion, not
-		 * as "text(name(varchar))" which the code below this point is
-		 * entirely capable of selecting.
+		 * want "text(varchar)" to be interpreted as a simple coercion, not as
+		 * "text(name(varchar))" which the code below this point is entirely
+		 * capable of selecting.
 		 *
-		 * "Trivial" coercions are ones that involve binary-compatible types and
-		 * ones that are coercing a previously-unknown-type literal constant
-		 * to a specific type.
+		 * We also treat a coercion of a previously-unknown-type literal
+		 * constant to a specific type this way.
 		 *
-		 * The reason we can restrict our check to binary-compatible coercions
-		 * here is that we expect non-binary-compatible coercions to have an
-		 * implementation function named after the target type. That function
-		 * will be found by normal lookup if appropriate.
+		 * The reason we reject COERCION_PATH_FUNC here is that we expect the
+		 * cast implementation function to be named after the target type.
+		 * Thus the function will be found by normal lookup if appropriate.
 		 *
-		 * NB: it's important that this code stays in sync with what coerce_type
+		 * The reason we reject COERCION_PATH_ARRAYCOERCE is mainly that you
+		 * can't write "foo[] (something)" as a function call.  In theory
+		 * someone might want to invoke it as "_foo (something)" but we have
+		 * never supported that historically, so we can insist that people
+		 * write it as a normal cast instead.  Lack of historical support is
+		 * also the reason for not considering composite-type casts here.
+		 *
+		 * NB: it's important that this code does not exceed what coerce_type
 		 * can do, because the caller will try to apply coerce_type if we
 		 * return FUNCDETAIL_COERCION.	If we return that result for something
 		 * coerce_type can't handle, we'll cause infinite recursion between
@@ -725,25 +762,34 @@ func_get_detail(List *funcname,
 		 */
 		if (nargs == 1 && fargs != NIL)
 		{
-			Oid			targetType;
-			TypeName   *tn = makeNode(TypeName);
+			Oid			targetType = FuncNameAsType(funcname);
 
-			tn->names = funcname;
-			tn->typmod = -1;
-			targetType = LookupTypeName(tn);
-			if (OidIsValid(targetType) &&
-				!ISCOMPLEX(targetType))
+			if (OidIsValid(targetType))
 			{
 				Oid			sourceType = argtypes[0];
 				Node	   *arg1 = linitial(fargs);
-				Oid			cfuncid;
+				bool		iscoercion;
 
-				if ((sourceType == UNKNOWNOID && IsA(arg1, Const)) ||
-					(find_coercion_pathway(targetType, sourceType,
-										   COERCION_EXPLICIT, &cfuncid) &&
-					 cfuncid == InvalidOid))
+				if (sourceType == UNKNOWNOID && IsA(arg1, Const))
 				{
-					/* Yup, it's a type coercion */
+					/* always treat typename('literal') as coercion */
+					iscoercion = true;
+				}
+				else
+				{
+					CoercionPathType cpathtype;
+					Oid			cfuncid;
+
+					cpathtype = find_coercion_pathway(targetType, sourceType,
+													  COERCION_EXPLICIT,
+													  &cfuncid);
+					iscoercion = (cpathtype == COERCION_PATH_RELABELTYPE ||
+								  cpathtype == COERCION_PATH_COERCEVIAIO);
+				}
+
+				if (iscoercion)
+				{
+					/* Treat it as a type coercion */
 					*funcid = InvalidOid;
 					*rettype = targetType;
 					*retset = false;
@@ -946,6 +992,33 @@ make_fn_arguments(ParseState *pstate,
 	}
 }
 
+/*
+ * FuncNameAsType -
+ *	  convenience routine to see if a function name matches a type name
+ *
+ * Returns the OID of the matching type, or InvalidOid if none.  We ignore
+ * shell types and complex types.
+ */
+static Oid
+FuncNameAsType(List *funcname)
+{
+	Oid			result;
+	Type		typtup;
+
+	typtup = LookupTypeName(NULL, makeTypeNameFromNameList(funcname), NULL);
+	if (typtup == NULL)
+		return InvalidOid;
+
+	if (((Form_pg_type) GETSTRUCT(typtup))->typisdefined &&
+		!OidIsValid(typeTypeRelid(typtup)))
+		result = typeTypeId(typtup);
+	else
+		result = InvalidOid;
+
+	ReleaseSysCache(typtup);
+	return result;
+}
+
 /*
  * ParseComplexProjection -
  *	  handles function calls with a single argument that is of complex type.
@@ -953,7 +1026,8 @@ make_fn_arguments(ParseState *pstate,
  *	  transformed expression tree.	If not, return NULL.
  */
 static Node *
-ParseComplexProjection(ParseState *pstate, char *funcname, Node *first_arg)
+ParseComplexProjection(ParseState *pstate, char *funcname, Node *first_arg,
+					   int location)
 {
 	TupleDesc	tupdesc;
 	int			i;
@@ -977,15 +1051,15 @@ ParseComplexProjection(ParseState *pstate, char *funcname, Node *first_arg)
 									 ((Var *) first_arg)->varno,
 									 ((Var *) first_arg)->varlevelsup);
 		/* Return a Var if funcname matches a column, else NULL */
-		return scanRTEForColumn(pstate, rte, funcname);
+		return scanRTEForColumn(pstate, rte, funcname, location);
 	}
 
 	/*
 	 * Else do it the hard way with get_expr_result_type().
 	 *
-	 * If it's a Var of type RECORD, we have to work even harder: we have to find
-	 * what the Var refers to, and pass that to get_expr_result_type. That
-	 * task is handled by expandRecordVariable().
+	 * If it's a Var of type RECORD, we have to work even harder: we have to
+	 * find what the Var refers to, and pass that to get_expr_result_type.
+	 * That task is handled by expandRecordVariable().
 	 */
 	if (IsA(first_arg, Var) &&
 		((Var *) first_arg)->vartype == RECORDOID)
@@ -1019,7 +1093,8 @@ ParseComplexProjection(ParseState *pstate, char *funcname, Node *first_arg)
  * helper routine for delivering "column does not exist" error message
  */
 static void
-unknown_attribute(ParseState *pstate, Node *relref, char *attname)
+unknown_attribute(ParseState *pstate, Node *relref, char *attname,
+				  int location)
 {
 	RangeTblEntry *rte;
 
@@ -1033,7 +1108,8 @@ unknown_attribute(ParseState *pstate, Node *relref, char *attname)
 		ereport(ERROR,
 				(errcode(ERRCODE_UNDEFINED_COLUMN),
 				 errmsg("column %s.%s does not exist",
-						rte->eref->aliasname, attname)));
+						rte->eref->aliasname, attname),
+				 parser_errposition(pstate, location)));
 	}
 	else
 	{
@@ -1044,18 +1120,21 @@ unknown_attribute(ParseState *pstate, Node *relref, char *attname)
 			ereport(ERROR,
 					(errcode(ERRCODE_UNDEFINED_COLUMN),
 					 errmsg("column \"%s\" not found in data type %s",
-							attname, format_type_be(relTypeId))));
+							attname, format_type_be(relTypeId)),
+					 parser_errposition(pstate, location)));
 		else if (relTypeId == RECORDOID)
 			ereport(ERROR,
 					(errcode(ERRCODE_UNDEFINED_COLUMN),
 			   errmsg("could not identify column \"%s\" in record data type",
-					  attname)));
+					  attname),
+					 parser_errposition(pstate, location)));
 		else
 			ereport(ERROR,
 					(errcode(ERRCODE_WRONG_OBJECT_TYPE),
 					 errmsg("column notation .%s applied to type %s, "
 							"which is not a composite type",
-							attname, format_type_be(relTypeId))));
+							attname, format_type_be(relTypeId)),
+					 parser_errposition(pstate, location)));
 	}
 }
 
@@ -1101,65 +1180,6 @@ func_signature_string(List *funcname, int nargs, const Oid *argtypes)
 									 nargs, argtypes);
 }
 
-/*
- * find_aggregate_func
- *		Convenience routine to check that a function exists and is an
- *		aggregate.
- *
- * Note: basetype is ANYOID if we are looking for an aggregate on
- * all types.
- */
-Oid
-find_aggregate_func(List *aggname, Oid basetype, bool noError)
-{
-	Oid			oid;
-	HeapTuple	ftup;
-	Form_pg_proc pform;
-
-	oid = LookupFuncName(aggname, 1, &basetype, true);
-
-	if (!OidIsValid(oid))
-	{
-		if (noError)
-			return InvalidOid;
-		if (basetype == ANYOID)
-			ereport(ERROR,
-					(errcode(ERRCODE_UNDEFINED_FUNCTION),
-					 errmsg("aggregate %s(*) does not exist",
-							NameListToString(aggname))));
-		else
-			ereport(ERROR,
-					(errcode(ERRCODE_UNDEFINED_FUNCTION),
-					 errmsg("aggregate %s(%s) does not exist",
-							NameListToString(aggname),
-							format_type_be(basetype))));
-	}
-
-	/* Make sure it's an aggregate */
-	ftup = SearchSysCache(PROCOID,
-						  ObjectIdGetDatum(oid),
-						  0, 0, 0);
-	if (!HeapTupleIsValid(ftup))	/* should not happen */
-		elog(ERROR, "cache lookup failed for function %u", oid);
-	pform = (Form_pg_proc) GETSTRUCT(ftup);
-
-	if (!pform->proisagg)
-	{
-		ReleaseSysCache(ftup);
-		if (noError)
-			return InvalidOid;
-		/* we do not use the (*) notation for functions... */
-		ereport(ERROR,
-				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
-				 errmsg("function %s(%s) is not an aggregate",
-					  NameListToString(aggname), format_type_be(basetype))));
-	}
-
-	ReleaseSysCache(ftup);
-
-	return oid;
-}
-
 /*
  * LookupFuncName
  *		Given a possibly-qualified function name and a set of argument types,
@@ -1194,6 +1214,27 @@ LookupFuncName(List *funcname, int nargs, const Oid *argtypes, bool noError)
 	return InvalidOid;
 }
 
+/*
+ * LookupTypeNameOid
+ *		Convenience routine to look up a type, silently accepting shell types
+ */
+static Oid
+LookupTypeNameOid(const TypeName *typename)
+{
+	Oid			result;
+	Type		typtup;
+
+	typtup = LookupTypeName(NULL, typename, NULL);
+	if (typtup == NULL)
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("type \"%s\" does not exist",
+						TypeNameToString(typename))));
+	result = typeTypeId(typtup);
+	ReleaseSysCache(typtup);
+	return result;
+}
+
 /*
  * LookupFuncNameTypeNames
  *		Like LookupFuncName, but the argument types are specified by a
@@ -1219,16 +1260,218 @@ LookupFuncNameTypeNames(List *funcname, List *argtypes, bool noError)
 	{
 		TypeName   *t = (TypeName *) lfirst(args_item);
 
-		argoids[i] = LookupTypeName(t);
-
-		if (!OidIsValid(argoids[i]))
-			ereport(ERROR,
-					(errcode(ERRCODE_UNDEFINED_OBJECT),
-					 errmsg("type \"%s\" does not exist",
-							TypeNameToString(t))));
-
+		argoids[i] = LookupTypeNameOid(t);
 		args_item = lnext(args_item);
 	}
 
 	return LookupFuncName(funcname, argcount, argoids, noError);
 }
+
+/*
+ * LookupAggNameTypeNames
+ *		Find an aggregate function given a name and list of TypeName nodes.
+ *
+ * This is almost like LookupFuncNameTypeNames, but the error messages refer
+ * to aggregates rather than plain functions, and we verify that the found
+ * function really is an aggregate.
+ */
+Oid
+LookupAggNameTypeNames(List *aggname, List *argtypes, bool noError)
+{
+	Oid			argoids[FUNC_MAX_ARGS];
+	int			argcount;
+	int			i;
+	ListCell   *lc;
+	Oid			oid;
+	HeapTuple	ftup;
+	Form_pg_proc pform;
+
+	argcount = list_length(argtypes);
+	if (argcount > FUNC_MAX_ARGS)
+		ereport(ERROR,
+				(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
+				 errmsg("functions cannot have more than %d arguments",
+						FUNC_MAX_ARGS)));
+
+	i = 0;
+	foreach(lc, argtypes)
+	{
+		TypeName   *t = (TypeName *) lfirst(lc);
+
+		argoids[i] = LookupTypeNameOid(t);
+		i++;
+	}
+
+	oid = LookupFuncName(aggname, argcount, argoids, true);
+
+	if (!OidIsValid(oid))
+	{
+		if (noError)
+			return InvalidOid;
+		if (argcount == 0)
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_FUNCTION),
+					 errmsg("aggregate %s(*) does not exist",
+							NameListToString(aggname))));
+		else
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_FUNCTION),
+					 errmsg("aggregate %s does not exist",
+							func_signature_string(aggname,
+												  argcount, argoids))));
+	}
+
+	/* Make sure it's an aggregate */
+	ftup = SearchSysCache(PROCOID,
+						  ObjectIdGetDatum(oid),
+						  0, 0, 0);
+	if (!HeapTupleIsValid(ftup))	/* should not happen */
+		elog(ERROR, "cache lookup failed for function %u", oid);
+	pform = (Form_pg_proc) GETSTRUCT(ftup);
+
+	if (!pform->proisagg)
+	{
+		ReleaseSysCache(ftup);
+		if (noError)
+			return InvalidOid;
+		/* we do not use the (*) notation for functions... */
+		ereport(ERROR,
+				(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+				 errmsg("function %s is not an aggregate",
+						func_signature_string(aggname,
+											  argcount, argoids))));
+	}
+
+	ReleaseSysCache(ftup);
+
+	return oid;
+}
+
+
+/*
+ * pg_get_expr() is a system function that exposes the expression
+ * deparsing functionality in ruleutils.c to users. Very handy, but it was
+ * later realized that the functions in ruleutils.c don't check the input
+ * rigorously, assuming it to come from system catalogs and to therefore
+ * be valid. That makes it easy for a user to crash the backend by passing
+ * a maliciously crafted string representation of an expression to
+ * pg_get_expr().
+ *
+ * There's a lot of code in ruleutils.c, so it's not feasible to add
+ * water-proof input checking after the fact. Even if we did it once, it
+ * would need to be taken into account in any future patches too.
+ *
+ * Instead, we restrict pg_rule_expr() to only allow input from system
+ * catalogs. This is a hack, but it's the most robust and easiest
+ * to backpatch way of plugging the vulnerability.
+ *
+ * This is transparent to the typical usage pattern of
+ * "pg_get_expr(systemcolumn, ...)", but will break "pg_get_expr('foo',
+ * ...)", even if 'foo' is a valid expression fetched earlier from a
+ * system catalog. Hopefully there aren't many clients doing that out there.
+ */
+void
+check_pg_get_expr_args(ParseState *pstate, Oid fnoid, List *args)
+{
+	Node	   *arg;
+
+	/* if not being called for pg_get_expr, do nothing */
+	if (fnoid != F_PG_GET_EXPR && fnoid != F_PG_GET_EXPR_EXT)
+		return;
+
+	/* superusers are allowed to call it anyway (dubious) */
+	if (superuser())
+		return;
+
+	/*
+	 * The first argument must be a Var referencing one of the allowed
+	 * system-catalog columns.  It could be a join alias Var or subquery
+	 * reference Var, though, so we need a recursive subroutine to chase
+	 * through those possibilities.
+	 */
+	Assert(list_length(args) > 1);
+	arg = (Node *) linitial(args);
+
+	if (!check_pg_get_expr_arg(pstate, arg, 0))
+		ereport(ERROR,
+				(errcode(ERRCODE_INSUFFICIENT_PRIVILEGE),
+				 errmsg("argument to pg_get_expr() must come from system catalogs")));
+}
+
+static bool
+check_pg_get_expr_arg(ParseState *pstate, Node *arg, int netlevelsup)
+{
+	if (arg && IsA(arg, Var))
+	{
+		Var		   *var = (Var *) arg;
+		RangeTblEntry *rte;
+		AttrNumber	attnum;
+
+		netlevelsup += var->varlevelsup;
+		rte = GetRTEByRangeTablePosn(pstate, var->varno, netlevelsup);
+		attnum = var->varattno;
+
+		if (rte->rtekind == RTE_JOIN)
+		{
+			/* Recursively examine join alias variable */
+			if (attnum > 0 &&
+				attnum <= list_length(rte->joinaliasvars))
+			{
+				arg = (Node *) list_nth(rte->joinaliasvars, attnum - 1);
+				return check_pg_get_expr_arg(pstate, arg, netlevelsup);
+			}
+		}
+		else if (rte->rtekind == RTE_SUBQUERY)
+		{
+			/* Subselect-in-FROM: examine sub-select's output expr */
+			TargetEntry *ste = get_tle_by_resno(rte->subquery->targetList,
+												attnum);
+			ParseState	mypstate;
+
+			if (ste == NULL || ste->resjunk)
+				elog(ERROR, "subquery %s does not have attribute %d",
+					 rte->eref->aliasname, attnum);
+			arg = (Node *) ste->expr;
+
+			/*
+			 * Recurse into the sub-select to see what its expr refers to.
+			 * We have to build an additional level of ParseState to keep in
+			 * step with varlevelsup in the subselect.
+			 */
+			MemSet(&mypstate, 0, sizeof(mypstate));
+			mypstate.parentParseState = pstate;
+			mypstate.p_rtable = rte->subquery->rtable;
+			/* don't bother filling the rest of the fake pstate */
+
+			return check_pg_get_expr_arg(&mypstate, arg, 0);
+		}
+		else if (rte->rtekind == RTE_RELATION)
+		{
+			switch (rte->relid)
+			{
+				case IndexRelationId:
+					if (attnum == Anum_pg_index_indexprs ||
+						attnum == Anum_pg_index_indpred)
+						return true;
+					break;
+
+				case AttrDefaultRelationId:
+					if (attnum == Anum_pg_attrdef_adbin)
+						return true;
+					break;
+
+				case ConstraintRelationId:
+					if (attnum == Anum_pg_constraint_conbin)
+						return true;
+					break;
+
+				case TypeRelationId:
+					if (attnum == Anum_pg_type_typdefaultbin)
+						return true;
+					break;
+			}
+		}
+	}
+
+	return false;
+}
