@@ -1,8 +1,10 @@
 /*
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2008, PostgreSQL Global Development Group
  * Portions Copyright (c) 1990 The Regents of the University of California.
  * All rights reserved.
  *
+ * $PostgreSQL: pgsql/src/backend/port/dynloader/netbsd.c,v 1.22.2.1 2008/03/05 21:20:47 alvherre Exp $
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -11,11 +13,7 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *	  notice, this list of conditions and the following disclaimer in the
  *	  documentation and/or other materials provided with the distribution.
- * 3. All advertising materials mentioning features or use of this software
- *	  must display the following acknowledgement:
- *		This product includes software developed by the University of
- *		California, Berkeley and its contributors.
- * 4. Neither the name of the University nor the names of its contributors
+ * 3. Neither the name of the University nor the names of its contributors
  *	  may be used to endorse or promote products derived from this software
  *	  without specific prior written permission.
  *
@@ -80,10 +78,10 @@ BSD44_derived_dlsym(void *handle, const char *name)
 	snprintf(error_message, sizeof(error_message),
 			 "dlsym (%s) failed", name);
 	return NULL;
-#elif defined(__ELF__)
-	return dlsym(handle, name);
 #else
 	void	   *vp;
+
+#ifndef __ELF__
 	char		buf[BUFSIZ];
 
 	if (*name != '_')
@@ -91,6 +89,7 @@ BSD44_derived_dlsym(void *handle, const char *name)
 		snprintf(buf, sizeof(buf), "_%s", name);
 		name = buf;
 	}
+#endif /* !__ELF__ */
 	if ((vp = dlsym(handle, (char *) name)) == NULL)
 		snprintf(error_message, sizeof(error_message),
 				 "dlsym (%s) failed", name);
